package eu.dnetlib.ariadneplus.reader;

import eu.dnetlib.ariadneplus.elasticsearch.BulkUpload;
import eu.dnetlib.ariadneplus.reader.ResourceManager;
import eu.dnetlib.ariadneplus.reader.json.ParseRDFJSON;
import org.eclipse.rdf4j.model.Model;
import org.eclipse.rdf4j.model.Statement;
import org.eclipse.rdf4j.query.*;
import org.eclipse.rdf4j.repository.Repository;
import org.eclipse.rdf4j.repository.RepositoryConnection;
import org.eclipse.rdf4j.repository.manager.RemoteRepositoryManager;
import org.eclipse.rdf4j.rio.RDFFormat;
import org.eclipse.rdf4j.rio.RDFWriter;
import org.eclipse.rdf4j.rio.Rio;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.List;

@Service
public class RunSPARQLQueryService {

    @Value("${sparql.query:undefined}")
    private String query;

    @Value("${repository.url:undefined}")
    private String repository_url;

    private RepositoryConnection connection;
    private RemoteRepositoryManager manager;
    private Repository repository;

    private ParseRDFJSON parser;
    private ResourceManager resourceManager;
    private BulkUpload bulkUpload;

    private static String username = null;
	private static String pwd = null;
	private static String graphDBUrl = null;
	private static String graphDBRepository = null;

	public void setupConnection(String username, String pwd, String graphDbUrl, String graphDbRepository) {
		setUsername(username);
		setPwd(pwd);
		setGraphDBUrl(graphDbUrl);
		setGraphDBRepository(graphDbRepository);
	}

	private void openConnection(){
    	manager = new RemoteRepositoryManager(getGraphDBUrl());
		manager.init();
		manager.setUsernameAndPassword(getUsername(), getPwd());
		repository = manager.getRepository(getGraphDBRepository());
        connection =  repository.getConnection();
        
    }

    private void closeConnection(){
        connection.close();
        repository.shutDown();
        manager.shutDown();
    }

	public String executeMultipleQueryGraph(String queryTemplate, List<String> recordIds, String datasource, String collectionId){
		if (queryTemplate==null)
			return null;
		final String selectQueryTemplate = queryTemplate.replaceAll("%datasource", datasource).replaceAll("%collectionId", collectionId);
		recordIds.forEach(recordId -> {
			executeQueryGraph(selectQueryTemplate, recordId);
		});
		return "ok";
	}

    private String executeQueryGraph(String selectQueryTemplate, String recordId){
    	query = selectQueryTemplate.replaceAll("%record", recordId);
    	System.out.println("execute query "+query);
		openConnection();
        StringWriter recordWriter = null;
        Model resultsModel = null;
        String jsonRecord = null;
        try {
        	System.out.println("Start connection Time: "+Calendar.getInstance().getTime().toString());
            GraphQuery graphQuery = connection.prepareGraphQuery(QueryLanguage.SPARQL, query);
            GraphQueryResult graphQueryResult = graphQuery.evaluate();
            resultsModel = QueryResults.asModel(graphQueryResult);
            graphQueryResult.close();
            System.out.println("End   connection Time: "+Calendar.getInstance().getTime().toString());
            System.out.println("count statements: " + resultsModel.size());
			if (resultsModel.size()==0) {
				return "noresult";
			}
            recordWriter = new StringWriter();
			RDFWriter rdfRecordWriter = Rio.createWriter(RDFFormat.RDFJSON, recordWriter);
		    Rio.write(resultsModel, rdfRecordWriter);
			System.out.println("RDF > json record: "+recordWriter.toString());
			parser.parse(recordWriter.toString());
			resourceManager.manage(parser);
			bulkUpload.index(resourceManager);
        } catch(Exception e){
            e.printStackTrace();
        } finally{
            closeConnection();
            if (resultsModel!=null) {
            	resultsModel.clear();
            }
        }
        return jsonRecord;
    }

    private void dumpModel(Model model) {
    	System.out.print(" [ dump model ] ");
		for (Statement stmt: model) {
			System.out.println(stmt.toString());
		}
    }

	public ParseRDFJSON getParser() {
		return parser;
	}

	
	public void setParser(ParseRDFJSON parser) {
		this.parser = parser;
	}

	
	public ResourceManager getResourceManager() {
		return resourceManager;
	}

	
	public void setResourceManager(ResourceManager resourceManager) {
		this.resourceManager = resourceManager;
	}

	
	public BulkUpload getBulkUpload() {
		return bulkUpload;
	}

	
	public void setBulkUpload(BulkUpload bulkUpload) {
		this.bulkUpload = bulkUpload;
	}

	public static String getUsername() {
		return username;
	}

	public static String getPwd() {
		return pwd;
	}

	public static String getGraphDBUrl() {
		return graphDBUrl;
	}

	public static String getGraphDBRepository() {
		return graphDBRepository;
	}

	public static void setUsername(String username) {
		RunSPARQLQueryService.username = username;
	}

	public static void setPwd(String pwd) {
		RunSPARQLQueryService.pwd = pwd;
	}

	public static void setGraphDBUrl(String graphDBUrl) {
		RunSPARQLQueryService.graphDBUrl = graphDBUrl;
	}

	public static void setGraphDBRepository(String graphDBRepository) {
		RunSPARQLQueryService.graphDBRepository = graphDBRepository;
	}

	public List<String> selectRecordIds(){
		query = "PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>\n" +
				"select * \n" +
				"from  <https://ariadne-infrastructure.eu/api_________::ariadne_plus::ads::271>\n" +
				"where { \n" +
				"\t?recordId rdf:type <https://www.ariadne-infrastructure.eu/resource/ao/cat/1.1/AO_Individual_Data_Resource> .\n" +
				"} \n";
		if (query.equals("undefined"))
			return null;
//		query = query.replaceAll("%s", "<https://ariadne-infrastructure.eu/aocat/Resource/02E4F4B5-24B7-3AD7-B460-CFA8B1F0BD1F>");
		openConnection();
		String jsonRecord = null;
		List<String> recordIds = new ArrayList<>();
		try {
			System.out.println("Start connection Time: "+Calendar.getInstance().getTime().toString());
			TupleQuery selectQuery = connection.prepareTupleQuery(QueryLanguage.SPARQL, query);
			TupleQueryResult selectQueryResult = selectQuery.evaluate();
			int counter = 0;
			while (selectQueryResult.hasNext()) {
				BindingSet recordSet = selectQueryResult.next();
				org.eclipse.rdf4j.model.Value recordIdValue = recordSet.getValue("recordId");
				System.out.println(recordIdValue.stringValue());
				recordIds.add(recordIdValue.stringValue());
				counter++;
			}
			System.out.println("Total records: "+counter);
			System.out.println("End   connection Time: "+Calendar.getInstance().getTime().toString());
		} catch(Exception e){
			e.printStackTrace();
		} finally{
			closeConnection();
		}
		return recordIds;
	}
}
